from typing import Any, Callable, Dict, Sequence, Type, Union
from unittest.mock import Mock

from ninja.constants import NOT_SET, NOT_SET_TYPE
from ninja.testing.client import NinjaResponse
from ninja_extra import ControllerBase
from ninja_extra.testing.client import NinjaExtraClientBase

from easy.main import EasyAPI


class EasyAPIClientBase(NinjaExtraClientBase):
    def __init__(
        self,
        controller_class: Union[Type[ControllerBase], Type],
        auth: Union[
            Sequence[Callable[..., Any]], Callable[..., Any], NOT_SET_TYPE, None
        ] = NOT_SET,
        api_cls: Union[Type[EasyAPI], Type] = EasyAPI,
    ) -> None:
        api = api_cls(auth=auth)
        assert hasattr(controller_class, "get_api_controller"), "Not a valid object"
        controller_ninja_api_controller = controller_class.get_api_controller()
        assert controller_ninja_api_controller
        controller_ninja_api_controller.set_api_instance(api)
        self._urls_cache = list(controller_ninja_api_controller.urls_paths(""))
        super(NinjaExtraClientBase, self).__init__(api)


class EasyTestClient(EasyAPIClientBase):
    async def _call(self, func: Callable, request: Mock, kwargs: Dict) -> NinjaResponse:
        return NinjaResponse(await func(request, **kwargs))



import logging
from typing import Optional

from ninja_extra import ControllerBase

from easy.controller.meta import CrudAPI, CrudApiMetaclass
from easy.services import BaseService

logger = logging.getLogger(__name__)


class CrudAPIController(ControllerBase, CrudAPI, metaclass=CrudApiMetaclass):
    """For Client facing APIs"""

    def __init__(self, service: Optional["BaseService"] = None):
        super().__init__(service=service)  # type: ignore



import json
import logging
import uuid
from abc import ABCMeta
from typing import Any, Optional, Tuple, Type, Union

from django.db import models
from django.http import HttpRequest
from ninja import ModelSchema
from ninja_extra import ControllerBase, http_delete, http_get, http_patch, http_put
from ninja_extra.pagination import paginate

from easy.domain.orm import CrudModel
from easy.services import BaseService
from easy.utils import copy_func

logger = logging.getLogger(__name__)


class CrudAPI(CrudModel):
    # Never add type note to service, it will cause injection error
    def __init__(self, service=None):  # type: ignore
        if not service:
            self.service = BaseService(model=self.model)
        else:
            self.service = service
        super().__init__(model=self.model)

    # Define Controller APIs for auto generation
    async def get_obj(self, request: HttpRequest, id: int) -> Any:
        """
        GET /{id}
        Retrieve a single Object
        """
        return await self.service.get_obj(id)

    async def del_obj(self, request: HttpRequest, id: int) -> Any:
        """
        DELETE /objects/{id}
        Delete a single Object
        """
        return await self.service.del_obj(id)

    @paginate
    async def get_objs(
        self, request: HttpRequest, maximum: int = None, filters: str = None
    ) -> Any:
        """
        GET /get_all
        Retrieve multiple Object
        """
        if filters:
            return await self.service.get_objs(maximum, **json.loads(filters))
        return await self.service.get_objs(maximum)

    @paginate
    async def filter_objs(
        self, request: HttpRequest, filters: Union[str, bytes]
    ) -> Any:
        """
        GET /filter/?filters={filters_dict}
        Filter Objects
        """
        return await self.service.filter_objs(**json.loads(filters))

    @paginate
    async def filter_exclude_objs(self, filters: Union[str, bytes]) -> Any:
        """
        GET /filter_exclude/?filters={filters_dict}
        Filter exclude Objects
        """
        return await self.service.filter_exclude_objs(**json.loads(filters))

    # async def bulk_create_objs(self, request):
    #     """
    #     POST /bulk_create
    #     Create multiple Object
    #     """
    #     return await self.service.bulk_create_objs()
    #
    # async def recover_obj(self, request):
    #     """
    #     PATCH /{id}/recover
    #     Recover one Object
    #     """
    #     return await self.service.recover_obj()


class CrudApiMetaclass(ABCMeta):
    def __new__(mcs, name: str, bases: Tuple[Type[Any], ...], attrs: dict) -> Any:
        # Get configs from Meta
        temp_base: Type = type.__new__(type, "object", (), {})
        temp_cls: Type = super(CrudApiMetaclass, mcs).__new__(
            mcs, name, (temp_base,), attrs
        )
        temp_opts: ModelOptions = ModelOptions(getattr(temp_cls, "Meta", None))
        opts_fields_exclude: Optional[Union[str]] = temp_opts.exclude
        opts_fields: Optional[Union[str]] = temp_opts.fields
        opts_model: Optional[Type[models.Model]] = temp_opts.model
        opts_recursive: Optional[Union[bool]] = temp_opts.recursive

        base_cls_attrs = {
            "get_obj": http_get("/", summary="Get")(
                copy_func(CrudAPI.get_obj)  # type: ignore
            ),
            "del_obj": http_delete("/", summary="Delete")(
                copy_func(CrudAPI.del_obj)  # type: ignore
            ),
            "get_all": http_get("/get_all", summary="Get All")(
                copy_func(CrudAPI.get_objs)  # type: ignore
            ),
            "filter_objs": http_get("/filter", summary="Filter")(
                copy_func(CrudAPI.filter_objs)  # type: ignore
            ),
            "filter_exclude_objs": http_get(
                "/filter_exclude", summary="Filter exclude"
            )(
                copy_func(CrudAPI.filter_exclude_objs)  # type: ignore
            ),
        }

        if opts_model:

            class DataSchema(ModelSchema):
                class Config:
                    model = opts_model
                    if opts_fields_exclude:
                        model_exclude = opts_fields_exclude
                    else:
                        if opts_fields == ["__all__"]:
                            model_fields = "__all__"
                        else:
                            model_fields = opts_fields if opts_fields else "__all__"
                    recursive = temp_opts.recursive

            async def add_obj(  # type: ignore
                self, request: HttpRequest, data: DataSchema
            ) -> Any:
                """
                PUT /
                Create a single Object
                """
                return await self.service.add_obj(**data.dict())

            async def patch_obj(  # type: ignore
                self, request: HttpRequest, id: int, data: DataSchema
            ) -> Any:
                """
                PATCH /?id={id}
                Update a single field for a Object
                """
                return await self.service.patch_obj(id=id, payload=data.dict())

            DataSchema.__name__ = (
                f"{opts_model.__name__}__AutoSchema({str(uuid.uuid4())[:4]})"
            )

            setattr(CrudAPI, "add_obj", classmethod(add_obj))
            setattr(CrudAPI, "patch_obj", classmethod(patch_obj))
            base_cls_attrs.update(
                {
                    "patch_obj": http_patch("/", summary="Patch/Update")(
                        copy_func(CrudAPI.patch_obj)  # type: ignore
                    ),
                    "add_obj": http_put("/", summary="Create")(
                        copy_func(CrudAPI.add_obj)  # type: ignore
                    ),
                }
            )

        new_base: Type = type.__new__(
            type, name, (ControllerBase, CrudAPI), base_cls_attrs
        )
        new_cls: Type = super(CrudApiMetaclass, mcs).__new__(
            mcs, name, (new_base,), attrs
        )

        new_cls.model = opts_model
        new_cls.fields_exclude = opts_fields_exclude
        new_cls.fields = opts_fields
        new_cls.recursive = opts_recursive

        return new_cls


class ModelOptions:
    def __init__(self, options: object = None):
        self.model: Optional[Type[models.Model]] = getattr(options, "model", None)
        self.service: Optional[Any] = getattr(options, "service", None)
        self.fields: Optional[Union[str]] = getattr(options, "fields", None)
        self.exclude: Optional[Union[str]] = getattr(options, "exclude", None)
        self.recursive: Optional[Union[bool]] = getattr(options, "recursive", False)



from ninja_extra.operation import AsyncOperation

from easy import EasyAPI

api_admin_v1 = EasyAPI()
api_admin_v1.auto_create_admin_controllers()


def test_auto_generate_admin_api():
    assert len(api_admin_v1._routers) == 5  # default + 3 models
    path_names = []
    controllers = []
    for path, rtr in api_admin_v1._routers:
        path_names.append(path)
        controllers.append(str(rtr))
        for path_ops in rtr.path_operations.values():
            for op in path_ops.operations:
                assert isinstance(op, AsyncOperation)
                assert op.api is api_admin_v1

    assert "/demo_app/category" in path_names
    assert "/demo_app/client" in path_names
    assert "/demo_app/event" in path_names
    assert "/demo_app/type" in path_names

    assert "CategoryAdminAPIController" in controllers
    assert "EventAdminAPIController" in controllers
    assert "ClientAdminAPIController" in controllers
    assert "TypeAdminAPIController" in controllers


def test_auto_generation_settings(settings):
    settings.AUTO_ADMIN_EXCLUDE_APPS = ["tests.demo_app"]
    api_admin_v2 = EasyAPI()
    api_admin_v2.auto_create_admin_controllers()
    assert len(api_admin_v2._routers) == 1

    settings.AUTO_ADMIN_ENABLED_ALL_APPS = False
    settings.AUTO_ADMIN_EXCLUDE_APPS = []
    settings.AUTO_ADMIN_INCLUDE_APPS = ["tests.none_existing_app"]
    api_admin_v3 = EasyAPI()
    api_admin_v3.auto_create_admin_controllers()
    assert len(api_admin_v3._routers) == 1



import logging
from typing import Type, Union

from django.db import models
from ninja_extra import ControllerBase, api_controller

from easy.controller.base import CrudAPIController
from easy.permissions import AdminSitePermission

logger = logging.getLogger(__name__)


class AdminClass(object):
    auto_import = True


def create_admin_controller(
    model: models.Model, app_name: str
) -> Union[Type[ControllerBase], Type]:
    """Create AdminAPI class dynamically, permission class set to AdminSitePermission"""
    model_name = model.__name__  # type:ignore
    Meta = type(
        "Meta",
        (object,),
        {"model": model, "fields": "__all__"},
    )

    class_name = f"{model_name}AdminAPIController"

    auto_cls = type.__new__(
        type,
        class_name,
        (
            CrudAPIController,
            AdminClass,
        ),
        {
            "Meta": Meta,
        },
    )

    logger.debug(f"Creating Admin API for f{model.__dict__.items()}")
    return api_controller(
        f"/{app_name}/{model_name.lower()}",
        tags=[f"{model_name} AdminAPI"],
        permissions=[AdminSitePermission],
    )(auto_cls)



import logging
from typing import Optional, Type

from django.db import models

from easy.domain import BaseDomain
from easy.services.crud import CrudService
from easy.services.permission import PermissionService

logger = logging.getLogger(__name__)


class BaseService(CrudService, PermissionService):
    def __init__(
        self,
        model: models.Model,
        biz: Optional[Type[BaseDomain]] = None,
    ):
        self.biz = biz
        if self.biz:
            self.model = self.biz.model
        else:
            self.model = model
        super().__init__(model=self.model)



import json
from datetime import datetime, timedelta

import django
import pytest
from asgiref.sync import sync_to_async

from tests.demo_app.controllers import AutoGenCrudAPIController, EventSchema
from tests.demo_app.models import Category, Client, Event, Type

dummy_data = dict(
    title="AsyncAdminAPIEvent",
    start_date=str(datetime.now().date()),
    end_date=str((datetime.now() + timedelta(days=5)).date()),
)


@pytest.mark.skipif(django.VERSION < (3, 1), reason="requires django 3.1 or higher")
@pytest.mark.django_db
class TestAutoCrudAdminAPI:
    async def test_crud_default_get_all(self, transactional_db, easy_api_client):
        client = easy_api_client(AutoGenCrudAPIController)

        object_data = dummy_data.copy()
        object_data.update(title=f"{object_data['title']}_get_all")

        event = await sync_to_async(Event.objects.create)(**object_data)
        type = await sync_to_async(Type.objects.create)(name="Type")
        category = await sync_to_async(Category.objects.create)(
            title="Category for Unit Testings"
        )
        client_a = await sync_to_async(Client.objects.create)(
            name="Client A for Unit Testings", key="A"
        )
        client_b = await sync_to_async(Client.objects.create)(
            name="Client B for Unit Testings", key="B"
        )
        event.category = category
        event.type = type
        await sync_to_async(event.save)()
        await sync_to_async(event.owner.set)([client_a, client_b])

        response = await client.get(
            "/get_all", query=dict(maximum=100, filters=json.dumps(dict(id__gte=1)))
        )
        assert response.status_code == 200

        data = response.json().get("data")
        assert data[0]["title"] == "AsyncAdminAPIEvent_get_all"
        assert data[0]["type"] == type.id

        event_schema = json.loads(EventSchema.from_orm(event).json())
        assert event_schema["start_date"] == data[0]["start_date"]

        response = await client.get(
            "/get_all",
            query=dict(
                maximum=100,
            ),
        )
        assert response.status_code == 200

        data = response.json().get("data")
        assert data[0]["title"] == "AsyncAdminAPIEvent_get_all"

        event_schema = json.loads(EventSchema.from_orm(event).json())
        assert event_schema["start_date"] == data[0]["start_date"]

        response = await client.get("/get_all")
        assert response.status_code == 200

        data = response.json().get("data")
        assert data[0]["title"] == "AsyncAdminAPIEvent_get_all"

        event_schema = json.loads(EventSchema.from_orm(event).json())
        assert event_schema["start_date"] == data[0]["start_date"]

    async def test_crud_default_get_delete(self, transactional_db, easy_api_client):
        client = easy_api_client(AutoGenCrudAPIController)

        object_data = dummy_data.copy()
        object_data.update(title=f"{object_data['title']}_get")

        event = await sync_to_async(Event.objects.create)(**object_data)

        response = await client.get(
            f"/?id={event.id}",
        )
        assert response.status_code == 200

        data = response.json().get("data")
        assert data["title"] == "AsyncAdminAPIEvent_get"

        event_schema = json.loads(EventSchema.from_orm(event).json())
        assert event_schema["end_date"] == data["end_date"]

        await client.delete(
            f"/?id={event.id}",
        )

        response = await client.get(
            f"/?id={event.id}",
        )
        assert response.status_code == 200
        assert response.json().get("data") == {}

    async def test_crud_default_create(self, transactional_db, easy_api_client):
        client = easy_api_client(AutoGenCrudAPIController)

        client_c = await sync_to_async(Client.objects.create)(
            name="Client C for Unit Testings", key="C"
        )

        client_d = await sync_to_async(Client.objects.create)(
            name="Client D for Unit Testings", key="D"
        )

        object_data = dummy_data.copy()
        object_data.update(title=f"{object_data['title']}_create")
        object_data.update(owner=[client_c.id, client_d.id])

        response = await client.put(
            "/", json=object_data, content_type="application/json"
        )
        assert response.status_code == 200

        event_id = response.json().get("data")["id"]

        response = await client.get(
            f"/?id={event_id}",
        )
        assert response.status_code == 200
        assert response.json().get("data")["title"] == "AsyncAdminAPIEvent_create"

    async def test_crud_default_patch(self, transactional_db, easy_api_client):
        client = easy_api_client(AutoGenCrudAPIController)

        object_data = dummy_data.copy()
        event = await sync_to_async(Event.objects.create)(**object_data)

        response = await client.get(
            f"/?id={event.id}",
        )
        assert response.status_code == 200
        assert response.json().get("data")["title"] == f"{object_data['title']}"

        client_e = await sync_to_async(Client.objects.create)(
            name="Client E for Unit Testings", key="E"
        )

        client_f = await sync_to_async(Client.objects.create)(
            name="Client F for Unit Testings", key="F"
        )

        new_data = dict(
            id=event.id,
            title=f"{object_data['title']}_patch",
            start_date=str((datetime.now() + timedelta(days=10)).date()),
            end_date=str((datetime.now() + timedelta(days=20)).date()),
            owner=[client_e.id, client_f.id],
        )

        response = await client.patch(
            f"/?id={event.id}", json=new_data, content_type="application/json"
        )

        assert response.status_code == 200
        assert response.json().get("data")["id"] == event.id
        assert response.json().get("data")["created"] is False

        response = await client.get(
            f"/?id={event.id}",
        )
        assert response.status_code == 200
        assert response.json().get("data")["title"] == "AsyncAdminAPIEvent_patch"
        assert response.json().get("data")["start_date"] == str(
            (datetime.now() + timedelta(days=10)).date()
        )
        assert response.json().get("data")["end_date"] == str(
            (datetime.now() + timedelta(days=20)).date()
        )

    async def test_default_filter(self, transactional_db, easy_api_client):
        object_data = dummy_data.copy()
        object_data.update(title=f"{object_data['title']}_filter")

        event = await sync_to_async(Event.objects.create)(**object_data)

        client = easy_api_client(AutoGenCrudAPIController)

        response = await client.get(
            "/filter", query=dict(filters=json.dumps(dict(id__gte=1)))
        )
        assert response.status_code == 200

        data = response.json().get("data")
        assert data[0]["title"] == "AsyncAdminAPIEvent_filter"

        event_schema = json.loads(EventSchema.from_orm(event).json())
        assert event_schema["end_date"] == data[0]["end_date"]

    async def test_crud_filter_exclude(self, transactional_db, easy_api_client):
        object_data = dummy_data.copy()
        object_data.update(title=f"{object_data['title']}_exclude")

        event = await sync_to_async(Event.objects.create)(**object_data)

        client = easy_api_client(AutoGenCrudAPIController)

        response = await client.get(
            "/filter_exclude", query=dict(filters=json.dumps(dict(id__lt=1)))
        )
        assert response.status_code == 200

        data = response.json().get("data")
        assert data[0]["title"] == "AsyncAdminAPIEvent_exclude"

        event_schema = json.loads(EventSchema.from_orm(event).json())
        assert event_schema["end_date"] == data[0]["end_date"]



from typing import Optional, Union

from ninja_extra import status
from ninja_extra.exceptions import APIException


class BaseAPIException(APIException):
    """
    API Exception
    """

    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
    default_detail = "There is an exception, please try again later."

    def __init__(
        self,
        detail: Optional[str] = None,
        code: Optional[Union[int]] = None,
    ) -> None:
        super(BaseAPIException, self).__init__(detail, code)
        if detail:
            self.default_detail = detail
        if code:
            self.status_code = code


class APIAuthException(BaseAPIException):
    """
    API Auth Exception
    """

    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = "Unauthorized"



from typing import Any, Dict, List, Tuple

from django.db import models

from easy.response import BaseApiResponse


def is_model_instance(data: Any) -> bool:
    return isinstance(data, models.Model)


def is_queryset(data: Any) -> bool:
    return isinstance(data, models.query.QuerySet)


def is_one_relationship(data: Any) -> bool:
    return isinstance(data, models.ForeignKey) or isinstance(data, models.OneToOneRel)


def is_many_relationship(data: Any) -> bool:
    return (
        isinstance(data, models.ManyToManyRel)
        or isinstance(data, models.ManyToManyField)
        or isinstance(data, models.ManyToOneRel)
    )


def is_paginated(data: Any) -> bool:
    return isinstance(data, dict) and isinstance(
        data.get("items", None), models.query.QuerySet
    )


def is_base_response(data: Any) -> bool:
    """Check if it is already good or a BaseApiResponse"""
    return isinstance(data, BaseApiResponse) and data.get("data", None) is not None


def serialize_base_response(data: Any) -> BaseApiResponse:
    return BaseApiResponse(data=data, message=None)


def serialize_model_instance(
    obj: models.Model, referrers: Any = tuple()
) -> Dict[Any, Any]:
    """Serializes Django model instance to dictionary"""
    out = {}
    for field in obj._meta.get_fields():
        # TODO: Add OneToOne relationship??
        if is_one_relationship(field):
            out.update(serialize_foreign_key(obj, field, referrers + (obj,)))

        elif is_many_relationship(field):
            out.update(serialize_many_relationship(obj, referrers + (obj,)))

        else:
            out.update(serialize_value_field(obj, field))
    return out


def serialize_queryset(
    data: models.query.QuerySet, referrers: Tuple[Any, ...] = tuple()
) -> List[Dict[Any, Any]]:
    """Serializes Django Queryset to dictionary"""
    return [serialize_model_instance(obj, referrers) for obj in data]


def serialize_foreign_key(
    obj: models.Model, field: Any, referrers: Any = tuple()
) -> Dict[Any, Any]:
    """Serializes foreign key field of Django model instance"""
    if not hasattr(obj, field.name):
        return {field.name: None}
    related_instance = getattr(obj, field.name)
    if related_instance is None:
        return {field.name: None}
    if related_instance in referrers:
        return {}
    return {field.name: getattr(related_instance, "pk")}
    # TODO: recursive
    # return {field.name: serialize_model_instance(related_instance, referrers)}


def serialize_many_relationship(
    obj: models.Model, referrers: Any = tuple()
) -> Dict[Any, Any]:
    """
    Serializes many relationship (ManyToMany, ManyToOne) of Django model instance
    """
    if not hasattr(obj, "_prefetched_objects_cache"):
        return {}
    out = {}
    for k, v in obj._prefetched_objects_cache.items():  # type: ignore
        field_name = k if hasattr(obj, k) else k + "_set"
        if v:
            out[field_name] = serialize_queryset(v, referrers + (obj,))
        else:
            out[field_name] = []
    return out


def serialize_value_field(obj: models.Model, field: Any) -> Dict[Any, Any]:
    """Serializes regular 'jsonable' field (Char, Int, etc.) of Django model instance"""
    if field.name in ["password"]:
        return {}
    return {field.name: getattr(obj, field.name)}



import logging
from typing import Any, Tuple

from django.db import models
from django.db.models.query import QuerySet
from ninja_extra.shortcuts import get_object_or_none

from easy.domain.serializers import is_many_relationship
from easy.response import BaseApiResponse

logger = logging.getLogger(__name__)


class CrudModel(object):
    def __init__(self, model: models.Model):
        self.model = model

    def __get_fields(self, payload: dict) -> Tuple[dict, dict]:
        m2m_fields = {}
        local_fields = {}
        for field in payload.keys():
            model_field = self.model._meta.get_field(field)
            if is_many_relationship(model_field):
                m2m_fields.update({field: payload[field]})
            else:
                local_fields.update({field: payload[field]})
        return local_fields, m2m_fields

    # Define BASE CRUD
    def _crud_add_obj(self, **payload: dict) -> Any:
        local_fields, m2m_fields = self.__get_fields(payload)
        obj = self.model.objects.create(**local_fields)
        if m2m_fields:
            for field, value in m2m_fields.items():
                if value and isinstance(value, list):
                    m2m_f = getattr(obj, field)
                    m2m_f.set(value)
        return obj

    def _crud_del_obj(self, id: int) -> "BaseApiResponse":
        obj = get_object_or_none(self.model, id=id)
        if obj:
            self.model.objects.filter(id=id).delete()
            return BaseApiResponse({"Detail": "Deleted."})
        else:
            return BaseApiResponse(
                {"Detail": "Not found."}, message="Not found."
            )  # pragma: no cover

    def _crud_update_obj(self, id: int, payload: dict) -> "BaseApiResponse":
        local_fields, m2m_fields = self.__get_fields(payload)
        try:
            obj, created = self.model.objects.update_or_create(
                id=id, defaults=local_fields
            )
        except Exception as e:  # pragma: no cover
            logger.error(f"Crud_update Error - {e}", exc_info=True)
            return BaseApiResponse(message="Failed")
        if m2m_fields:
            for field, value in m2m_fields.items():
                if value:
                    m2m_f = getattr(obj, field)
                    m2m_f.set(value)
        return BaseApiResponse({"id": obj.id, "created": created})

    def _crud_get_obj(self, id: int) -> Any:
        return get_object_or_none(self.model, id=id)

    def _crud_get_objs_all(self, maximum: int = None, **filters: Any) -> Any:
        """
        CRUD: get maximum amount of records, with filters support
        Args:
            maximum:
            filters: {"field_name__lte", 1}

        Returns: qs

        """
        qs = None
        if filters:
            try:
                qs = self.model.objects.filter(**filters)
            except Exception as e:  # pragma: no cover
                logger.error(e)
        elif maximum:
            qs = self.model.objects.all()[:maximum]
        else:
            qs = self.model.objects.all()
        return qs

    def _crud_filter(self, **kwargs: Any) -> QuerySet:
        return self.model.objects.filter(**kwargs)

    def _crud_filter_exclude(self, **kwargs: Any) -> QuerySet:
        return self.model.objects.all().exclude(**kwargs)


class BaseOrm(CrudModel):
    ...
